

# Cell
import open3d as o3d
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap, LinearSegmentedColormap
import yaml
import math
from queue import Queue, PriorityQueue
plt.rcParams["figure.figsize"] = (16, 16)

# Cell
import yaml
import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt
import math

# Cell
import yaml
import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt
import math

# Cell
import yaml
import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt
import math

# Cell
import yaml
import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt
import math

# Cell
import yaml
import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt
import math

# Cell
import yaml
import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt
import math

# Cell
import yaml
import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt
import math

# Cell
import open3d as o3d
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap, LinearSegmentedColormap
import pylab as plot
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import yaml
import math
import os
from queue import Queue, PriorityQueue
import random
plt.rcParams["figure.figsize"] = (16, 16)


# AUTOGENERATED! DO NOT EDIT! File to edit: 13_PokingBot_training_multi_machine.ipynb (unless otherwise specified).

__all__ = ['CURL_NAVI_GYM_V1']

# Cell
import socket
import _thread
import pickle
import time
import os
import sys
import signal
import subprocess
import numpy as np
import yaml
import random
import gym
from gym import error, spaces, utils
from gym.utils import seeding
import math
from math import e

# Cell

from .CurriculumSchedulerMultiMachine import CurriculumSchedulerMultiMachine
class CURL_NAVI_GYM_V1(gym.Env):
    def __init__(self):
        with open('Data/IP_ENV.yaml', 'r') as f: self.IP_ENV = yaml.load(f, Loader=yaml.CLoader)
        # Spaces
        self.observation_space = spaces.Box(high=100, low=0, shape=( (4+4)*241+10*3, ) ) # laser max range=100
        self.action_space = spaces.Box(low=np.array([0, -1, 0]), high=np.array([1, 1, 1]), dtype=np.float32)
        # state info
        self.info = {'laser_shape': (4+4, 241),'goal_shape': (10, 3)} # 4 frames laser range, 4 frames laser label, 10 frames pose

        self.observation_state = np.zeros(((4+4)*241+10*3,), dtype=np.float32)
        self.action_state = np.array([0,0,0], dtype=np.float32)
        self.reward = float(0)
        self.done_state = (False,False) # done & collision -> (True,False),
                                        # done & goal reached -> (True,True)
        self.reset_gazebo_model = False
        self.start_goal_pair = {'goal': [0,0,1], 'start': [0,0,1]}
        self.action_scale = {'angular': 0.8, 'linear': 1.5, 'ur5': 0.5}
        self.have_received_data = False
        self.step_count = 0

        self.scheduler = CurriculumSchedulerMultiMachine('Data/curriculum.yaml')
        self.socket = socket.socket()         # Create a socket object
        host = '0.0.0.0' # Get local machine name
        port = 9453      # Reserve a port for your service.
        print('Server started!')
        print('Waiting for clients...')

        self.socket.bind((host, port))        # Bind to the port
        self.socket.listen(15)                 # Now wait for client connection.
        _thread.start_new_thread(self.socket_server_thread,(1,))


    def socket_server_thread(self,_):
        while True:
            c, addr = self.socket.accept()     # Establish connection with client
            _thread.start_new_thread(self.on_new_client,(c,addr))

    def on_new_client(self, clientsocket, addr):
        while True:
            env = self.IP_ENV[addr[0]]

            if not (env == self.scheduler.current_task_env):
                msg = clientsocket.recv(256000)
                pkg = {}
                clientsocket.send(pickle.dumps(pkg))
                continue

            # receive
            time.sleep(0.2) # sleeping is very important !!!
            msg = clientsocket.recv(256000)
            try:
                pack_from_client = pickle.loads(msg)
                self.observation_state = pack_from_client["observation_state"]
                self.reward = pack_from_client["reward"]
                self.done_state = pack_from_client["done_state"]
                self.have_received_data = True
            except Exception as e:
                pass

            #send
            pack_to_client = { "action_state": self.action_state, \
                               "reset_gazebo_model": self.reset_gazebo_model, \
                               "start_goal_pair": self.start_goal_pair , \
                               "action_scale": self.action_scale }
            clientsocket.send(pickle.dumps(pack_to_client))
            if self.reset_gazebo_model : self.reset_gazebo_model = False

        clientsocket.close()


    def reset(self):
        self.start_goal_pair, self.action_scale = self.scheduler.update(self.done_state[1])
        self.reset_gazebo_model = True
        self.step_count = 0
        return self.observation_state

    def step(self, action):
        while not self.have_received_data:
            pass
        self.action_state = action
        self.step_count += 1
        self.have_received_data = False
        return self.observation_state, self.reward, self.done_state[0], self.info




# AUTOGENERATED! DO NOT EDIT! File to edit: 00_Evaluation.ipynb (unless otherwise specified).

__all__ = ['evaluation']

# Cell
import math
import numpy as np
import pickle as pkl
import csv
import time

class evaluation():
    def __init__(self):
        self.count = 0
        self.epoch = 0
        self.iteration = 5
        self.collision = 0
        self.robot_radius = 3.5
        self.stop = False
        self.goal = None
        self.goal_distance = 0
        self.record = np.zeros([self.iteration])
        self.record_collision = np.zeros([self.iteration])
        self.collision_states = False
        self.start_time = 0
        self.time = 0
    def get_goal(self, pt_list, index):
        """Get goal from list"""
        self.goal = pt_list[index]

    def odom_calculation(self, robot_pose):
        """determine whether the robot has reach the goal"""
        self.goal_distance = self.get_distance(robot_pose, self.goal)
        if self.goal_distance < self.robot_radius and self.stop == False:
            self.record[self.epoch] = 1
            self.record_collision[self.epoch] = self.collision
            self.epoch += 1
        elif self.collision > 50:
            self.record[self.epoch] = 0
            self.record_collision[self.epoch] = self.collision
            self.epoch += 1
        if self.epoch == self.iteration and self.stop == False:
            record_name = 'goal_navigation_cave_reshape_wamv.csv'
            fileObject = open("./" + record_name, 'w')
            writer = csv.writer(fileObject)
            writer.writerow(self.record)
            writer.writerow(self.record_collision)
            fileObject.close()
            print("done")
            self.stop = True

    def collision_calculation(self, msg):
        """calculate collision times"""
        if self.collision_states == True:
            if msg.states == [] and self.count > 1000:
                self.collision_states = False
            else:
                self.count += 1
        elif msg.states != [] and self.count == 0:
            self.collision_states = True
            self.collision += 1
            print("collsion time: ", self.collision)
        else:
            self.collision_states = False
            self.count = 0

    def get_distance(self, p1, p2):
        """calculate the distance from robot pose to goal"""
        return math.sqrt((p1[0] - p2[0])**2 + (p1[1] - p2[1])**2)




# AUTOGENERATED! DO NOT EDIT! File to edit: 04_DoorGym.ipynb (unless otherwise specified).

__all__ = ['get_validation_args', 'get_training_args', 'eval_print', 'onpolicy_validation', 'mujoco_timestep',
           'onpolicy_main']

# Cell
import os
import sys
import time
import pickle
from collections import deque
import numpy as np
import argparse
import torch
import torch.nn as nn
from tensorboardX import SummaryWriter
import time

from a2c_ppo_acktr.visionmodel import VisionModelXYZ, VisionModel
from a2c_ppo_acktr import algo, utils
from a2c_ppo_acktr.envs import make_vec_envs, VecPyTorch
from a2c_ppo_acktr.model import Policy
from a2c_ppo_acktr.storage import RolloutStorage
from a2c_ppo_acktr.utils import get_vec_normalize, get_render_func

# Cell
def get_validation_args():
    """Create arguments for validation.

    This funcion will build some arguments, like load model, environment and rendered engine.

    Args:
        None.

    Returns:
        The all parameters for validation.

    """
    parser = argparse.ArgumentParser(description='DoorGym validation')
    parser.add_argument(
        '--seed', type=int, default=1, help='random seed (default: 1)')
    parser.add_argument(
        '--log-interval',
        type=int,
        default=10,
        help='log interval, one log per n updates (default: 10)')
    parser.add_argument(
        '--env-name',
        default='doorenv-v0',
        help='environment to train on (default: PongNoFrameskip-v4)')
    parser.add_argument(
        '--non-det',
        action='store_true',
        default=False,
        help='whether to use a non-deterministic policy')
    parser.add_argument(
        '--load-name',
        type=str,
        default='',
        help='which model to load')
    parser.add_argument(
        '--eval',
        action='store_true',
        default=False,
        help="Measure the opening ratio among 100 trials")
    parser.add_argument(
        '--render',
        action='store_true',
        default=False,
        help="force rendering")
    parser.add_argument(
        '--knob-noisy',
        action='store_true',
        default=False,
        help='add noise to knob position to resemble the noise from the visionnet')
    parser.add_argument(
        '--visionnet-input',
        action="store_true",
        default=False,
        help='Use vision net for knob position estimation')
    parser.add_argument(
        '--unity',
        action="store_true",
        default=False,
        help='Use unity for an input of a vision net')
    parser.add_argument(
        '--port',
        type=int,
        default=1050,
        help='Unity connection port (Only for off-policy)')
    parser.add_argument(
        '--visionmodel-path',
        type=str,
        default="./trained_visionmodel/",
        help='load the replay buffer')
    parser.add_argument(
        '--world-path',
        type=str,
        default="/u/home/urakamiy/pytorch-a2c-ppo-acktr-gail/random_world/world/pull_floatinghook",
        help='load the vision network model')
    parser.add_argument(
        '--pos-control',
        action="store_true",
        default=False,
        help='use pos control')
    parser.add_argument(
        '--step-skip',
        type=int,
        default=4,
        help='number of step skip in pos control')
    args = parser.parse_args()

    args.det = not args.non_det

    return args

# Cell
def get_training_args():
    """Create arguments for training.

    This funcion will build some arguments, like algorithm, environment and learning rate.

    Args:
        None.

    Returns:
        The all parameters for training.

    """
    parser = argparse.ArgumentParser(description='DoorGym training')
    parser.add_argument(
        '--algo', default='ppo', help='algorithm to use: a2c | ppo | sac | td3')
    parser.add_argument(
        '--lr', type=float, default=1e-3, help='learning rate (default: 1e-3)')
    parser.add_argument(
        '--eps',
        type=float,
        default=1e-5,
        help='RMSprop optimizer epsilon (default: 1e-5)')
    parser.add_argument(
        '--alpha',
        type=float,
        default=0.99,
        help='RMSprop optimizer apha (default: 0.99)')
    parser.add_argument(
        '--gamma',
        type=float,
        default=0.99,
        help='discount factor for rewards (default: 0.99)')
    parser.add_argument(
        '--use-gae',
        action='store_false',
        default=True,
        help='use generalized advantage estimation')
    parser.add_argument(
        '--gae-lambda',
        type=float,
        default=0.95,
        help='gae lambda parameter (default: 0.95)')
    parser.add_argument(
        '--entropy-coef',
        type=float,
        default=0.01,
        help='entropy term coefficient (default: 0.01)')
    parser.add_argument(
        '--value-loss-coef',
        type=float,
        default=0.5,
        help='value loss coefficient (default: 0.5)')
    parser.add_argument(
        '--max-grad-norm',
        type=float,
        default=0.5,
        help='max norm of gradients (default: 0.5)')
    parser.add_argument(
        '--seed', type=int, default=1, help='random seed (default: 1)')
    parser.add_argument(
        '--cuda-deterministic',
        action='store_true',
        default=False,
        help="sets flags for determinism when using CUDA (potentially slow!)")
    parser.add_argument(
        '--num-processes',
        type=int,
        default=8,
        help='how many training CPU processes to use (default: 8)')
    parser.add_argument(
        '--num-steps',
        type=int,
        default=4096,
        help='number of forward steps (default: 4096)')
    parser.add_argument(
        '--ppo-epoch',
        type=int,
        default=10,
        help='number of ppo epochs (default: 10)')
    parser.add_argument(
        '--num-mini-batch',
        type=int,
        default=256,
        help='number of batches for ppo (default: 256)')
    parser.add_argument(
        '--clip-param',
        type=float,
        default=0.2,
        help='ppo clip parameter (default: 0.2)')
    parser.add_argument(
        '--log-interval',
        type=int,
        default=1,
        help='log interval, one log per n updates (default: 1)')
    parser.add_argument(
        '--save-interval',
        type=int,
        default=5,
        help='save interval, one save per n updates (default: 5)')
    parser.add_argument(
        '--eval-interval',
        type=int,
        default=20,
        help='eval interval, one eval per n updates (default: None)')
    parser.add_argument(
        '--num-env-steps',
        type=int,
        default=10e8,
        help='number of environment steps to train (default: 10e6)')
    parser.add_argument(
        '--env-name',
        default='doorenv-v0',
        help='environment to train on (default: doorenv-v0)')
    parser.add_argument(
        '--log-dir',
        default='./logs/',
        help='directory to save agent logs (default: ./logs/)')
    parser.add_argument(
        '--params-log-dir',
        default='./params_logs/',
        help='directory to save params logs (default: ./params_logs/)')
    parser.add_argument(
        '--save-dir',
        default='./trained_models/',
        help='directory to save agent logs (default: ./trained_models/)')
    parser.add_argument(
        '--no-cuda',
        action='store_true',
        default=False,
        help='disables CUDA training')
    parser.add_argument(
        '--use-proper-time-limits',
        action='store_false',
        default=True,
        help='compute returns taking into account time limits')
    parser.add_argument(
        '--recurrent-policy',
        action='store_true',
        default=False,
        help='use a recurrent policy')
    parser.add_argument(
        '--use-linear-lr-decay',
        action='store_true',
        default=False,
        help='use a linear schedule on the learning rate')
    parser.add_argument(
        '--save-name',
        type=str,
        default="test",
        help='name for changing the log and model name')
    parser.add_argument(
        '--knob-noisy',
        action='store_true',
        default=False,
        help='add noise to knob position to resemble the noise from the visionnet')
    parser.add_argument(
        '--obs-noisy',
        action='store_true',
        default=False,
        help='add noise to entire observation signal')
    parser.add_argument(
        '--pretrained-policy-load',
        type=str,
        default=None,
        help='which pretrained model to load')
    parser.add_argument(
        '--replaybuffer-load',
        type=str,
        default=None,
        help='load the replay buffer')
    parser.add_argument(
        '--visionmodel-path',
        type=str,
        default="./trained_visionmodel/",
        help='load the vision network model')
    parser.add_argument(
        '--world-path',
        type=str,
        default="/home/demo/doorgym/world_generator/world/pull_floatinghook",
        help='load the vision network model')
    parser.add_argument(
        '--val-path',
        type=str,
        default=None,
        help='load the vision network model')
    parser.add_argument(
        '--visionnet-input',
        action="store_true",
        default=False,
        help='Use vision net for knob position estimation')
    parser.add_argument(
        '--unity',
        action="store_true",
        default=False,
        help='Use unity for an input of a vision net')
    parser.add_argument(
        '--port',
        type=int,
        default=1050,
        help='port number to connect to Unity. (Only for SAC)')
    parser.add_argument(
        '--step-skip',
        type=int,
        default=4,
        help='number of step skip in pos control')
    parser.add_argument(
        '--pos-control',
        action="store_true",
        default=False,
        help='Turn on pos control')

    args = parser.parse_args()

    args.cuda = not args.no_cuda and torch.cuda.is_available()

    assert args.algo in ['a2c', 'ppo', 'sac', 'td3']
    if args.recurrent_policy:
        assert args.algo in ['a2c', 'ppo'], \
            'Recurrent policy is not implemented for sac'

    if args.unity:
        assert args.visionnet_input, \
            'Visionnet_input should be True when Unity is True'

    return args

# Cell
mujoco_timestep = 0.02
def eval_print(dooropen_counter, counter, start_time, total_time):
    """Print metric funcion.

    This function will show Average open door time and open door rate.

    Args:
        dooropen_counter :  Success open door conuter.
        counter : The total conuter.
        start_time : Start execute open door time.
        total_time : Total open door time.

    Return :
        opening_rate : Success open door rate.
        opening_timeavg : Average open door time.

    """

    opening_rate = dooropen_counter/counter *100
    if dooropen_counter != 0:
        opening_timeavg = total_time/dooropen_counter
    else:
        opening_timeavg = -1
    print("opening rate {}%. Average time to open is {}.".format(opening_rate, opening_timeavg))
    print("took {}sec to evaluate".format( int(time.mktime(time.localtime())) - start_time ))

    return opening_rate, opening_timeavg

def onpolicy_validation(
                seed,
                env_name,
                det,
                load_name,
                evaluation,
                render,
                knob_noisy,
                visionnet_input,
                env_kwargs,
                actor_critic=None,
                verbose=True,
                pos_control=True,
                step_skip=4):

    """Validation on-policy method

    This function will execute validation trained model.

    Args:
        seed : Setting seed to random.
        env_name : Test Environment.
        det : Whether to use a deterministic policy.
        load_name : Input trained model.
        evaluation : True is evaluation, False is validation.
        render : Whether to render this environment.
        knob_noisy : Whether to add noise into vision.
        visionnet_input : Use vision net to get knob position.
        env_kwargs : Other environment arguments.
        actor_critic : Model network arhitecture.
        verbose : Whether to show opening rate and timeavg.
        pos_control : Whether to use pos control.
        step_skip : The number of step skip in pos control.

    Return:
        opening_rate : Success open door rate.
        opening_timeavg : Average open door time.

    """

    env = make_vec_envs(
        env_name,
        seed + 1000,
        1,
        None,
        None,
        device='cuda:0',
        allow_early_resets=False,
        env_kwargs=env_kwargs,)

    env_obj = env.venv.venv.envs[0].env.env
    if env_name.find('door')<=-1:
        env_obj.unity = None

    render_func = get_render_func(env)
    if evaluation and not render:
        render_func = None

    if env_kwargs['visionnet_input']:
        visionmodel = VisionModelXYZ()
        visionmodel_path = args.visionmodel_path
        if not os.path.isfile(visionmodel_path):
            raise RuntimeError("=> no checkpoint found at '{}'" .format(visionmodel_path))
        checkpoint = torch.load(visionmodel_path)
        visionmodel.load_state_dict(checkpoint['state_dict'])
        best_pred = checkpoint['best_pred']
        print("=> loaded checkpoint '{}' (epoch {}), best pred {}".format(visionmodel_path, checkpoint['epoch'], best_pred))

    if not actor_critic:
            actor_critic, ob_rms = torch.load(load_name)
    actor_critic = actor_critic.eval()
    if env_kwargs['visionnet_input'] and env_name.find('doorenv')>-1:
        actor_critic.visionmodel = visionmodel
        actor_critic.visionnet_input = env_obj.visionnet_input
    actor_critic.to("cuda:0")

    if env_name.find('doorenv')>-1:
        actor_critic.nn = env_obj.nn

    recurrent_hidden_states = torch.zeros(1,actor_critic.recurrent_hidden_state_size)
    masks = torch.zeros(1, 1)

    full_obs = env.reset()
    initial_state = full_obs[:,:env.action_space.shape[0]]

    if env_name.find('doorenv')>-1 and env_obj.visionnet_input:
        obs = actor_critic.obs2inputs(full_obs, 0)
    else:
        if knob_noisy:
            satulation = 100.
            sdv = torch.tensor([3.440133806003181, 3.192113342496682,  1.727412865751099]) /satulation  #Vision SDV for arm
            noise = torch.distributions.Normal(torch.tensor([0.0, 0.0, 0.0]), sdv).sample().cuda()
            noise *= min(1., 100/satulation)
            obs[:,-3:] += noise
        else:
            obs = full_obs

    if render_func is not None:
        render_func('human')

    start_time = int(time.mktime(time.localtime()))

    i=0
    epi_step = 0
    total_time = 0
    epi_counter = 1
    dooropen_counter = 0
    door_opened = False
    test_num = 100

    while True:
        with torch.no_grad():
            value, action, _, recurrent_hidden_states = actor_critic.act(
                obs, recurrent_hidden_states, masks, deterministic=det)
        next_action = action

        if pos_control:
            if i%(512/step_skip-1)==0: current_state = initial_state
            next_action = current_state + next_action
            for kk in range(step_skip):
                full_obs, reward, done, infos = env.step(next_action)

            current_state = full_obs[:,:env.action_space.shape[0]]
        else:
            for kk in range(step_skip):
                full_obs, reward, done, infos = env.step(next_action)


        if env_name.find('doorenv')>-1 and env_obj.visionnet_input:
            obs = actor_critic.obs2inputs(full_obs, 0)
        else:
            if knob_noisy:
                satulation = 100.
                sdv = torch.tensor([3.440133806003181, 3.192113342496682,  1.727412865751099]) /satulation  #Vision SDV for arm
                noise = torch.distributions.Normal(torch.tensor([0.0, 0.0, 0.0]), sdv).sample().cuda()
                noise *= min(1., 100/satulation)
                obs[:,-3:] += noise
            else:
                obs = full_obs

        masks.fill_(0.0 if done else 1.0)

        if render_func is not None:
            render_func('human')

        i+=1
        epi_step += 1

        if env_name.find('doorenv')>-1:
            if not door_opened and abs(env_obj.get_doorangle())>=0.2:
                dooropen_counter += 1
                opening_time = epi_step/(1.0/mujoco_timestep)*step_skip
                if verbose:
                    print("door opened! opening time is {}".format(opening_time))
                total_time += opening_time
                door_opened = True

        if env_name.find('Fetch')>-1:
            if not door_opened and infos[0]['is_success']==1:
                dooropen_counter += 1
                opening_time = epi_step/(1.0/mujoco_timestep)*step_skip
                if verbose:
                    print("Reached destenation! Time is {}".format(opening_time))
                total_time += opening_time
                door_opened = True

        if evaluation:
            if i%(512/step_skip-1)==0:
                if env_obj.unity:
                    env_obj.close()
                env = make_vec_envs(
                env_name,
                seed + 1000,
                1,
                None,
                None,
                device='cuda:0',
                allow_early_resets=False,
                env_kwargs=env_kwargs,)

                if render:
                    render_func = get_render_func(env)
                env_obj = env.venv.venv.envs[0].env.env
                if env_name.find('doorenv')<=-1:
                    env_obj.unity = None
                env.reset()
                if verbose:
                    print("{} ep end >>>>>>>>>>>>>>>>>>>>>>>>".format(epi_counter))
                    eval_print(dooropen_counter, epi_counter, start_time, total_time)
                epi_counter += 1
                epi_step = 0
                door_opened = False


        if i>=512/step_skip*test_num:
            if verbose:
                print( "dooropening counter:",dooropen_counter, " epi counter:", epi_counter)
                eval_print(dooropen_counter, epi_counter-1, start_time, total_time)
            break

    opening_rate, opening_timeavg = eval_print(dooropen_counter, epi_counter-1, start_time, total_time)

    return opening_rate, opening_timeavg

# Cell
def onpolicy_main(args):
    """On-policy training function

    This function will set training environment, including parameters setting, load model, training and evaluation.

    Args:
        Args : The all training parameters.

    Return :
        None.

    """
    print("onpolicy main")

    # torch cuda setting
    torch.manual_seed(args.seed)
    torch.cuda.manual_seed_all(args.seed)
    if args.cuda and torch.cuda.is_available() and args.cuda_deterministic:
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True
    torch.set_num_threads(1)
    device = torch.device("cuda:0" if args.cuda else "cpu")

    # tensorboard
    summary_name = args.log_dir + '{0}_{1}'
    writer = SummaryWriter(summary_name.format(args.env_name, args.save_name))

    # create env key word arguments
    env_kwargs = dict(port = args.port,
                    visionnet_input = args.visionnet_input,
                    unity = args.unity,
                    world_path = args.world_path,
                    pos_control = args.pos_control)

    env_kwargs_val = env_kwargs.copy()
    if args.val_path: env_kwargs_val['world_path'] = args.val_path

    # Make vector env
    envs = make_vec_envs(args.env_name,
                         args.seed,
                         args.num_processes,
                         args.gamma,
                         args.log_dir,
                         device,
                         False,
                         env_kwargs=env_kwargs,)

    # agly ways to access to the environment attirubutes
    if args.env_name.find('doorenv')>-1:
        if args.num_processes>1:
            visionnet_input = envs.venv.venv.visionnet_input
            nn = envs.venv.venv.nn
            env_name = envs.venv.venv.xml_path
        else:
            visionnet_input = envs.venv.venv.envs[0].env.env.env.visionnet_input
            nn = envs.venv.venv.envs[0].env.env.env.nn
            env_name = envs.venv.venv.envs[0].env.env.env.xml_path
        # state dimensions
        dummy_obs = np.zeros(23)
    else:
        dummy_obs = envs.observation_space
        visionnet_input = None
        nn = None

    # load pretrained
    if args.pretrained_policy_load:
        print("loading", args.pretrained_policy_load)
        actor_critic, ob_rms = torch.load(args.pretrained_policy_load)
    else:
        actor_critic = Policy(
            dummy_obs.shape,
            envs.action_space,
            base_kwargs={'recurrent': args.recurrent_policy})

    # vision network
    if visionnet_input:
        visionmodel = VisionModelXYZ()
        visionmodel_path = args.visionmodel_path
        if not os.path.isfile(visionmodel_path):
            raise RuntimeError("=> no checkpoint found at '{}'" .format(visionmodel_path))
        checkpoint = torch.load(visionmodel_path)
        visionmodel.load_state_dict(checkpoint['state_dict'])
        best_pred = checkpoint['best_pred']
        print("=> loaded checkpoint '{}' (epoch {}), best pred {}".format(visionmodel_path, checkpoint['epoch'], best_pred))
        actor_critic.visionmodel = visionmodel.eval()

    actor_critic.nn = nn
    actor_critic.to(device)

    #disable normalizer
    vec_norm = get_vec_normalize(envs)
    vec_norm.eval()

    # create agent
    if args.algo == 'a2c':
        agent = algo.A2C_ACKTR(
            actor_critic,
            args.value_loss_coef,
            args.entropy_coef,
            lr=args.lr,
            eps=args.eps,
            alpha=args.alpha,
            max_grad_norm=args.max_grad_norm)
    elif args.algo == 'ppo':
        agent = algo.PPO(
            actor_critic,
            args.clip_param,
            args.ppo_epoch,
            args.num_mini_batch,
            args.value_loss_coef,
            args.entropy_coef,
            lr=args.lr,
            eps=args.eps,
            max_grad_norm=args.max_grad_norm)

    # replay buffer
    rollouts = RolloutStorage(args.num_steps, args.num_processes,
                              dummy_obs.shape, envs.action_space,
                              actor_critic.recurrent_hidden_state_size)

    # get observation
    full_obs = envs.reset()
    initial_state = full_obs[:,:envs.action_space.shape[0]]

    # preprocessing
    if args.env_name.find('doorenv')>-1 and visionnet_input:
        obs = actor_critic.obs2inputs(full_obs, 0)
    else:
        if args.knob_noisy:
            satulation = 100.
            sdv = torch.tensor([3.440133806003181, 3.192113342496682, 1.727412865751099]) /100.  #Vision SDV for arm
            noise = torch.distributions.Normal(torch.tensor([0.0, 0.0, 0.0]), sdv).sample().cuda()
            noise *= min(1., 0/satulation)
            obs[:,-3:] += noise
        elif args.obs_noisy:
            sdv = torch.ones(obs.size(1))*0.03
            noise = torch.distributions.Normal(torch.zeros(sdv.size()), sdv).sample().cuda()
            obs[:] += noise
        else:
            obs = full_obs

    rollouts.obs[0].copy_(obs)
    rollouts.to(device)

    episode_rewards = deque(maxlen=10)

    start = time.time()
    num_updates = int(
        args.num_env_steps) // args.num_steps // args.num_processes

    # training
    for j in range(num_updates):

        if args.use_linear_lr_decay:
            # decrease learning rate linearly
            utils.update_linear_schedule(
                agent.optimizer, j, num_updates, args.lr)

        for step in range(args.num_steps):
            with torch.no_grad():
                value, action, action_log_prob, recurrent_hidden_states = actor_critic.act(
                    rollouts.obs[step], rollouts.recurrent_hidden_states[step],
                    rollouts.masks[step])
                next_action = action

            if args.pos_control:
                if step%(512/args.step_skip-1)==0: current_state = initial_state
                next_action = current_state + next_action
                for kk in range(args.step_skip):
                    full_obs, reward, done, infos = envs.step(next_action)

                current_state = full_obs[:,:envs.action_space.shape[0]]
            else:
                for kk in range(args.step_skip):
                    full_obs, reward, done, infos = envs.step(next_action)

            # convert img to obs if door_env and using visionnet
            if args.env_name.find('doorenv')>-1 and visionnet_input:
                obs = actor_critic.obs2inputs(full_obs, j)
            else:
                if args.knob_noisy:
                    satulation = 100.
                    sdv = torch.tensor([3.440133806003181, 3.192113342496682, 1.727412865751099]) /100.  #Vision SDV for arm
                    noise = torch.distributions.Normal(torch.tensor([0.0, 0.0, 0.0]), sdv).sample().cuda()
                    noise *= min(1., j/satulation)
                    obs[:,-3:] += noise
                elif args.obs_noisy:
                    sdv = torch.ones(obs.size(1))*0.03
                    noise = torch.distributions.Normal(torch.zeros(sdv.size()), sdv).sample().cuda()
                    obs[:] += noise
                else:
                    obs = full_obs

            for info in infos:
                if 'episode' in info.keys():
                    episode_rewards.append(info['episode']['r'])

            masks = torch.FloatTensor(
                [[0.0] if done_ else [1.0] for done_ in done])
            bad_masks = torch.FloatTensor(
                [[0.0] if 'bad_transition' in info.keys() else [1.0]
                 for info in infos])
            rollouts.insert(obs, recurrent_hidden_states, action,
                            action_log_prob, value, reward, masks, bad_masks)

        with torch.no_grad():
            next_value = actor_critic.get_value(
                rollouts.obs[-1], rollouts.recurrent_hidden_states[-1],
                rollouts.masks[-1]).detach()

        rollouts.compute_returns(next_value, args.use_gae, args.gamma,
                                 args.gae_lambda, args.use_proper_time_limits)

        value_loss, action_loss, dist_entropy = agent.update(rollouts)
        rollouts.after_update()

        # Get total number of timesteps
        total_num_steps = (j + 1) * args.num_processes * args.num_steps

        writer.add_scalar("Value loss", value_loss, j)
        writer.add_scalar("action loss", action_loss, j)
        writer.add_scalar("dist entropy loss", dist_entropy, j)
        writer.add_scalar("Episode rewards", np.mean(episode_rewards), j)

        # save for every interval-th episode or for the last epoch
        if (j % args.save_interval == 0
                or j == num_updates - 1) and args.save_dir != "":
            save_path = os.path.join(args.save_dir, args.algo)
            try:
                os.makedirs(save_path)
            except OSError:
                pass
            torch.save([
                actor_critic,
                getattr(utils.get_vec_normalize(envs), 'ob_rms', None)
            ], os.path.join(save_path, args.env_name + "_{}.{}.pt".format(args.save_name,j)))

        if j % args.log_interval == 0 and len(episode_rewards) > 1:
            end = time.time()
            print(
                "Updates {}, num timesteps {}, FPS {} \n Last {} training episodes: mean/median reward {:.1f}/{:.1f}, min/max reward {:.1f}/{:.1f}\n"
                .format(j, total_num_steps,
                        int(total_num_steps / (end - start)),
                        len(episode_rewards), np.mean(episode_rewards),
                        np.median(episode_rewards), np.min(episode_rewards),
                        np.max(episode_rewards), dist_entropy, value_loss,
                        action_loss))
        # evaluation
        if (args.eval_interval is not None and len(episode_rewards) > 1
                and j % args.eval_interval == 0):

            opening_rate, opening_timeavg = onpolicy_validation(
                                                seed=args.seed,
                                                env_name=args.env_name,
                                                det=True,
                                                load_name=args.save_name,
                                                evaluation=True,
                                                render=False,
                                                knob_noisy=args.knob_noisy,
                                                visionnet_input=args.visionnet_input,
                                                env_kwargs=env_kwargs_val,
                                                actor_critic=actor_critic,
                                                verbose=False,
                                                pos_control=args.pos_control,
                                                step_skip=args.step_skip)

            print("{}th update. {}th timestep. opening rate {}%. Average time to open is {}.".format(j, total_num_steps, opening_rate, opening_timeavg))
            writer.add_scalar("Opening rate per envstep", opening_rate, total_num_steps)
            writer.add_scalar("Opening rate per update", opening_rate, j)

        DR=True #Domain Randomization
        ################## for multiprocess world change ######################
        if DR:
            print("changing world")

            envs.close_extras()
            envs.close()
            del envs

            envs = make_vec_envs(args.env_name,
                        args.seed,
                        args.num_processes,
                        args.gamma,
                        args.log_dir,
                        device,
                        False,
                        env_kwargs=env_kwargs,)

            full_obs = envs.reset()
            if args.env_name.find('doorenv')>-1 and visionnet_input:
                obs = actor_critic.obs2inputs(full_obs, j)
            else:
                obs = full_obs
        #######################################################################


# AUTOGENERATED! DO NOT EDIT! File to edit: 02_pokingbot_training.ipynb (unless otherwise specified).

__all__ = ['PolicyNetwork', 'CriticNetwork', 'Trainer']

# Cell
from typing import Mapping, Sequence
from absl import app
from absl import flags
import acme
from acme import specs
from acme import types
from acme import wrappers
from acme.agents.tf import actors
from acme.agents.tf import d4pg
from acme.tf import networks
from acme.tf import utils as tf2_utils
from acme.tf import utils as tf2_utils
from acme.tf import networks
from acme import types
import dm_env
import gym
import numpy as np
import sonnet as snt
import sonnet as snt
import tensorflow as tf

# Cell
class PolicyNetwork(snt.Module):
    def __init__(self, action_spec):
        super(PolicyNetwork, self).__init__()

        # Get total number of action dimensions from action spec.
        output_dim = np.prod(action_spec.shape, dtype=int)

        self.conv = snt.Sequential([
            snt.Reshape([-1, 4]),
            snt.Conv1D(output_channels=32, kernel_shape=3, stride=1),
            tf.nn.relu,
            snt.Conv1D(output_channels=32, kernel_shape=3, stride=1),
            tf.nn.relu,
            snt.Flatten(),
        ])

        self.hidden = snt.Sequential([
            snt.Linear(512, w_init=tf.initializers.VarianceScaling(distribution='uniform', mode='fan_out', scale=0.333) ),
            tf.nn.elu,
            snt.Linear(512, w_init=tf.initializers.VarianceScaling(distribution='uniform', mode='fan_out', scale=0.333) ),
            tf.nn.elu,
            snt.Linear(256, w_init=tf.initializers.VarianceScaling(distribution='uniform', mode='fan_out', scale=0.333) ),
            tf.nn.elu,
            snt.Linear(256, w_init=tf.initializers.VarianceScaling(distribution='uniform', mode='fan_out', scale=0.333) ),
            tf.nn.elu,
        ])

        self.out = snt.Sequential([
            snt.Linear(
                output_dim, w_init=tf.initializers.VarianceScaling(1e-4)),
            networks.TanhToSpec(action_spec),
        ])

    def __call__(self, x):
        x = tf2_utils.batch_concat(x)

        laser, g= tf.split(x, [4*(241*2), 10*3], axis=1)
        laser = self.conv(laser)

        x = tf.concat([g, laser], 1)
        x = self.hidden(x)
        x = self.out(x)

        return x


class CriticNetwork(snt.Module):
    def __init__(self, vmin, vmax, num_atoms):
        super(CriticNetwork, self).__init__()

        self.conv = snt.Sequential([
            snt.Reshape([-1, 4]),
            snt.Conv1D(output_channels=32, kernel_shape=3, stride=1),
            tf.nn.relu,
            snt.Conv1D(output_channels=32, kernel_shape=3, stride=1),
            tf.nn.relu,
            snt.Flatten(),
        ])

        self.hidden = snt.Sequential([
            snt.Linear(512, w_init=tf.initializers.VarianceScaling(distribution='uniform', mode='fan_out', scale=0.333) ),
            tf.nn.elu,
            snt.Linear(512, w_init=tf.initializers.VarianceScaling(distribution='uniform', mode='fan_out', scale=0.333) ),
            tf.nn.elu,
            snt.Linear(256, w_init=tf.initializers.VarianceScaling(distribution='uniform', mode='fan_out', scale=0.333) ),
            tf.nn.elu,
            snt.Linear(256, w_init=tf.initializers.VarianceScaling(distribution='uniform', mode='fan_out', scale=0.333) ),
            tf.nn.elu,
        ])

        self.head = networks.DiscreteValuedHead(vmin, vmax, num_atoms)

    def __call__(self,
                 observation: types.NestedTensor,
                 action: types.NestedTensor) -> tf.Tensor:

        if hasattr(observation, 'dtype') and hasattr(action, 'dtype'):
            if observation.dtype != action.dtype:
                # Observation and action must be the same type for concat to work
                action = tf.cast(action, observation.dtype)

        x = tf2_utils.batch_concat(observation)
        action = tf2_utils.batch_concat(action)

        laser, g= tf.split(x, [4*(241*2), 10*3], axis=1)
        laser = self.conv(laser)

        x = tf.concat([g, laser, action], 1)
        x = self.hidden(x)
        x = self.head(x)

        return x


# Cell

class Trainer():
	def __init__(self, gym_id, pretrained=None):
		"""
		gym_id (String): The custom OpenAI gym environment have been registered. i.e. 'gym_poking:door-and-ball-v3'

		pretrained (String): While using a pretrained weight, the path of the specific weight, i.e.

								 log_home = '/home/argsubt/pokingbot-rl/acme_logs/'
								 log_id = '355a222e-7e90-11ec-9e19-04d4c401c495/'
								 snapshots_path = 'snapshots/policy'
								 checkpoints_path = 'checkpoints/d4pg_learner/'
								 ckpt_id = 'ckpt-409'
								 pretrained_weight = log_home + log_id + checkpoints_path + ckpt_id

		"""
		self.gym_id = gym_id
		self.specified_path = pretrained
	def __make_environment(self,
							task: str = 'gym_poking:curl-navi-gym-v0') -> dm_env.Environment:
		"""
		Creates an OpenAI Gym environment.
		Args:
			gym_id (String): The custom OpenAI gym environment have been registered.
		Returns:
			environment (env class)
		"""
		environment = gym.make(task)
		environment = wrappers.GymWrapper(environment)
		environment = wrappers.SinglePrecisionWrapper(environment)

		return environment

	def __make_networks(self,
						action_spec: specs.BoundedArray,
						vmin: float = -40.,
						vmax: float = 130.,
						num_atoms: int = 51,
						) -> Mapping[str, types.TensorTransformation]:
		"""
		Creates the networks used by the agent.
		vmin/vmax and num_atoms hyperparameters should be set to
		give the distributional critic a good dynamic range over possible discounted
		returns. Note that this is very different than the scale of immediate rewards.

		The vmin|vmax hyperparameters of the distributional critic may need tuning
		depending on your environment's rewards. A good rule of thumb is to
		set vmax to the discounted sum of the maximum instantaneous rewards
		for the maximum episode length;
		then set vmin to -vmax.

		Args:
			action_spec (specs.BoundedArray)
			vmin (float)
			vmax (float)
			num_atoms (int)
		Returns:
			policy_network (snt.Module)
			critic_network (snt.Module)
			observation_network (tf2_utils.batch_concat)
		"""

		observation_network = tf2_utils.batch_concat
		policy_network = PolicyNetwork(action_spec)
		critic_network = CriticNetwork(vmin, vmax, num_atoms)

		return {
			'policy': policy_network,
			'critic': critic_network,
			'observation': observation_network,
		}
	def main(self):
		"""
		Start running the train loop.
		"""
		environment = self.__make_environment(self.gym_id)
		environment_spec = specs.make_environment_spec(environment)
		agent_networks = self.__make_networks(environment_spec.actions)

		# change batch_size together with samples_per_insert
		# batch_size/samples_per_insert = 8
		# depends on the performance of workstation
		if self.specified_path == None:
			agent = d4pg.D4PG(
				environment_spec=environment_spec,
		        policy_network=agent_networks['policy'],
		        critic_network=agent_networks['critic'],
		        observation_network=agent_networks['observation'],
		        discount=0.99,
		        batch_size=512,
		        prefetch_size=4,
		        target_update_period=100,
		        min_replay_size=10000,
		        max_replay_size=200000,
		        # max_replay_size=1000000,
		        samples_per_insert=64.0,
		        n_step=5,
		        sigma=1.0,
		        clipping=True
			)
		else:
			agent = d4pg.D4PG(
				environment_spec=environment_spec,
		        policy_network=agent_networks['policy'],
		        critic_network=agent_networks['critic'],
		        observation_network=agent_networks['observation'],
		        discount=0.99,
		        batch_size=512,
		        prefetch_size=4,
		        target_update_period=100,
		        min_replay_size=10000,
		        max_replay_size=200000,
		        # max_replay_size=1000000,
		        samples_per_insert=64.0,
		        n_step=5,
		        sigma=1.0,
		        clipping=True,
				specified_path=self.specified_path
			)


		train_loop = acme.EnvironmentLoop(environment, agent, label='train_loop')
		train_loop.run(num_episodes=350000000)


# AUTOGENERATED! DO NOT EDIT! File to edit: 00_Evaluation_tools.ipynb (unless otherwise specified).

__all__ = ['list_to_queue', 'goal_buffer_handler', 'get_value', 'save_to_csv', 'get_distance']

# Cell
import math
import numpy as np
import pickle as pkl
import queue
import csv
import time

def list_to_queue(goal_list):
    """
    Args:
        goal_list: The goal pairs from our mission
    Return:
        points: transfer goal list to queue for handling goals
    """
    points = queue.Queue(maxsize=20)
    for i,pt in enumerate(goal_list):
        points.put(pt)
    return points

def goal_buffer_handler(point_queue):
    """
    Args:
        point_queue: goal point queue
    Return:
        current_goal: current goal for the robot to reach
    """
    current_goal = point_queue.get()
    return current_goal

def get_value(dict, target):
    """
    Args:
        dict: dictionary information from yaml file
        target: specific which path following task
    Return:
        v1: specific task information
    """
    for k, v in dict.items():
        for k1,v1 in v.items():
            if k1 == target:
                return v1

def save_to_csv(folder, name, record_information):
    """
    Args:
        folder: path
        name: file name
        record_information: recorded information
    """
    fileObject = open(folder + "/" + name, 'w')
    writer = csv.writer(fileObject)
    writer.writerow(record_information)
    fileObject.close()

def get_distance(p1, p2):
    """
    This function will return the distance from robot position to goal
    Args:
        p1: [x1,y1]
        p2: [x2,y2]
    Return:
        distance from p1 to p2
    """
    return math.sqrt((p1[0] - p2[0])**2 + (p1[1] - p2[1])**2)




# AUTOGENERATED! DO NOT EDIT! File to edit: 06_pokingbot_inference.ipynb (unless otherwise specified).

__all__ = []

# Cell
import yaml
import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt
import math


# AUTOGENERATED! DO NOT EDIT! File to edit: 03_diff_visualization.ipynb (unless otherwise specified).

__all__ = ['DifficultyMetrics', 'DifficultyVisualizer']

# Cell
class DifficultyMetrics:

  # map: C-space occupancy grid
  # path: list of points (row, col)
  # disp_radius: radius for dispersion
  def __init__(self, map, path, disp_radius):
    self.map = map
    self.rows = len(map)
    self.cols = len(map[0])
    self.axes = [(0, 1), (1, 1), (1, 0), (1, -1)] # vertical, horizontal, and 2 diagonals
    self.path = path
    self.radius = disp_radius

  # returns grid with the distance to closest obstacle at each point
  def closest_wall(self):
    dists = [[0 for i in range(self.cols)] for j in range(self.rows)]
    for r in range(self.rows):
      for c in range(self.cols):

        dists[r][c] = self._dist_closest_wall(r, c)

    return dists

  # returns grid with the average visibility at each point
  def avg_visibility(self):
    vis = [[0 for i in range(self.cols)] for j in range(self.rows)]
    for r in range(self.rows):
      for c in range(self.cols):
        vis[r][c] = self._avg_vis_cell(r, c)

    return vis

  # returns grid with the dispersion at each point
  # checks along 16 axes within the dispersion radius
  def dispersion(self):
    disp = [[0 for i in range(self.cols)] for j in range(self.rows)]
    for r in range(self.rows):
      for c in range(self.cols):
        disp[r][c] = self._cell_dispersion(r, c, self.radius)

    return disp

  # returns grid with the characteristic dimension at each point
  # characteristic dimension calculated in 2 directions for 4 axes
  def characteristic_dimension(self):
    cdr = [[0 for i in range(self.cols)] for j in range(self.rows)]
    for r in range(self.rows):
      for c in range(self.cols):
        if self.map[r][c] == 1:
          cdr[r][c] = 0

        cdr_min = self.rows + self.cols
        for axis in self.axes:
          cdr_min = min(cdr_min, self._distance(r, c, axis))

        cdr[r][c] = cdr_min

    return cdr

  # returns the distance along the axis in both directions, not including (r, c)
  def _distance(self, r, c, axis):
    if self.map[r][c] == 1:
      return -1

    # check axis in both directions
    reverse_axis = (axis[0] * -1, axis[1] * -1)
    dist = 0
    for move in [axis, reverse_axis]:
      r_curr = r
      c_curr = c

      # move in axis direction until an obstacle is found
      while self.map[r_curr][c_curr] != 1:
        r_curr += move[0]
        c_curr += move[1]

        if not self._in_map(r_curr, c_curr):
          break

        # add the distance traveled to the total
        if self.map[r_curr][c_curr] != 1:
          dist += math.sqrt(move[0] ** 2 + move[1] ** 2)

    return dist


  # a and b are tuples (row, col)
  def _dist_between_points(self, a, b):
    return math.sqrt((a[0] - b[0]) ** 2 + (a[1] - b[1]) ** 2)

  # path is list of points as tuples (row, col)
  # returns tortuosity calculated using the formula (arc length / chord length)
  def tortuosity(self):
    # calculate arc length
    arc_len = 0.0
    for i in range(1, len(self.path)):
      arc_len += self._dist_between_points(self.path[i - 1], self.path[i])

    # calculate chord length
    chord_len = self._dist_between_points(self.path[0], self.path[-1])
    return arc_len / chord_len

  # returns the cell dispersion within the radius
  def _cell_dispersion(self, r, c, radius):
    if self.map[r][c] == 1:
      return -1

    axes_wall = []
    # four cardinal, four diagonal, and one in between each (slope +- 1/2 or +-2)
    for move in [(0, 1), (1, 2), (1, 1), (2, 1), (1, 0), (2, -1), (1, -1), (1, -2), (0, -1), (-2, -1), (-1, -1), (-1, -2), (-1, 0), (-2, 1), (-1, 1), (-1, 2)]:
      count = 0
      wall = False
      r_curr = r
      c_curr = c
      while count < radius and not wall:
        r_curr += move[0]
        c_curr += move[1]

        if r_curr < 0 or r_curr >= self.rows or c_curr < 0 or c_curr >= self.cols:
          break

        if self.map[r_curr][c_curr] == 1:
          wall = True

        # count the in-between axes as two steps
        if move[0] == 2 or move[1] == 2:
          count += 2
        else:
          count += 1

      if wall:
        axes_wall.append(True)
      else:
        axes_wall.append(False)

    # count the number of changes in this cell's field of view
    change_count = 0
    for i in range(len(axes_wall)-1):
      if axes_wall[i] != axes_wall[i+1]:
        change_count += 1

    if axes_wall[0] != axes_wall[15]:
      change_count += 1

    return change_count

  # returns the average visibility at the point (r, c)
  def _avg_vis_cell(self, r, c):
    total_vis = 0.0
    num_axes = 0
    for r_move in [-1, 0, 1]:
      for c_move in [-1, 0, 1]:
        if r_move == 0 and c_move == 0:
          continue

        this_vis = 0
        r_curr = r
        c_curr = c
        wall_found = False
        while not wall_found:
          if not self._in_map(r_curr, c_curr):
            break

          if self.map[r_curr][c_curr] == 1:
            wall_found = True
          else:
            this_vis += math.sqrt((r_move ** 2) + (c_move ** 2))

          r_curr += r_move
          c_curr += c_move

        total_vis += this_vis
        num_axes += 1

    return total_vis / num_axes


  # bounds check
  def _in_map(self, r, c):
    return r >= 0 and r < self.rows and c >= 0 and c < self.cols

  # returns the distance to the closest obstacle at point (r, c)
  # returns 0 if self.map[r][c] is an obstacle, 1 if an adjacent non-diagonal cell is an obstacle, etc.
  def _dist_closest_wall(self, r, c):
    pq = PriorityQueue()
    first_wrapper = self.Wrapper(0, r, c)
    pq.put(first_wrapper)
    visited = {(r, c) : first_wrapper}

    while not pq.empty():
      point = pq.get()
      if self.map[point.r][point.c] == 1: # found an obstacle!
        return point.dist
      else:
        # enqueue all neighbors if they are in the map and have not been visited
        for row in range(point.r - 1, point.r + 2):
          for col in range(point.c - 1, point.c + 2):
            if self._in_map(row, col) and (row, col) not in visited:
              dist = math.sqrt((row - r) ** 2 + (col - c) ** 2)
              neighbor = self.Wrapper(dist, row, col)
              pq.put(neighbor)
              visited[(row, col)] = neighbor

    # in case the queue is empty before a wall is found (shouldn't happen),
    # the farthest a cell can be from a wall is half the board, since the top and bottom rows are all walls
    return (self.rows - 1) / 2

  # wrapper class for coordinates
  class Wrapper:

    def __init__(self, distance, row, col):
      self.dist = distance
      self.r = row
      self.c = col

    def __lt__(self, value):
      return self.dist < value.dist

  # returns a list of the unnormalized metrics, averaged over the path
  # metrics order: distance to closest wall, average visibility, dispersion,
  # characteristic dimension, and tortuosity
  def avg_all_metrics(self):
    result = []
    total = 0.0

    # closest wall
    total = 0.0
    for row, col in self.path:
      total += self._dist_closest_wall(row, col)
    result.append(total / len(self.path))

    # average visibility
    total = 0.0
    for row, col in self.path:
      total += self._avg_vis_cell(row, col)
    result.append(total / len(self.path))

    # dispersion
    total = 0.0
    for row, col in self.path:
      total += self._cell_dispersion(row, col, self.radius)
    result.append(total / len(self.path))

    # characteristic dimension
    total = 0.0
    char_dim_grid = self.characteristic_dimension()
    for row, col in self.path:
      total += char_dim_grid[row][col]
    result.append(total / len(self.path))

    # tortuosity
    tort = self.tortuosity()
    result.append(tort)

    return result

# Cell
class DifficultyVisualizer:
    def __init__(self, yaml_path, trajectory_yaml_list = []):
        self.yaml_path = yaml_path
        self.trajectory_yaml_list = trajectory_yaml_list
        with open(self.yaml_path, 'r') as stream:
            self.yaml_data = yaml.load(stream, Loader=yaml.CLoader)
        print(self.yaml_data)
        self.read_ply()


    def analyze_trajectory(self):
        for index, baseline in enumerate(self.trajectory_yaml_list):
            count = 0
            closest_wall = 0
            visibility = 0
            dispersion = 0
            characteristic_dimension = 0
            total_tortuosity = 0

            with open(baseline) as f: yaml_data = yaml.safe_load(f)
            for path in yaml_data:
                arc_len = 0.0
                for path_point_index, xy in enumerate(path):
                    c = int((xy[0] - self.origin_x)/self.resolution)
                    r = int((xy[1] - self.origin_y)/self.resolution)

                    if c >= self.map_height or c < 0 : continue
                    if r >= self.map_width or r < 0 : continue

                    closest_wall += self.total_closest_wall[r][c]
                    visibility += self.total_avg_visibility[r][c]
                    dispersion += self.total_dispersion[r][c]
                    characteristic_dimension += self.total_characteristic_dimension[r][c]
                    count += 1

                    # Tortuosity
                    if path_point_index==0: continue
                    arc_len += ((path[path_point_index][0] - path[path_point_index-1][0])**2 +
                                (path[path_point_index][1] - path[path_point_index-1][1])**2)**0.5
                chord_len = ((path[len(path)-1][0] - path[0][0])**2 +
                                (path[len(path)-1][1] - path[0][1])**2)**0.5
                tortuosity = arc_len / chord_len
                total_tortuosity += tortuosity


            print("------------------------- ",baseline, " -------------------------")
            print("avg closest_wall:", closest_wall/count)
            print("avg visibility:", visibility/count)
            print("avg dispersion:", dispersion/count)
            print("avg characteristic_dimension", characteristic_dimension/count)
            print("avg path tortuosity", total_tortuosity/len(yaml_data))


    def plot_trajectory(self):
        cmap = ListedColormap(['white', 'black', 'red', 'green'])
        vis_trajectory_map = self.map_bit.copy()
        for color_index,baseline in enumerate(self.trajectory_yaml_list):
            with open(baseline) as f: yaml_data = yaml.safe_load(f)

            for path in yaml_data:
                for xy in path:
                    c = int((xy[0] - self.origin_x)/self.resolution)
                    r = int((xy[1] - self.origin_y)/self.resolution)

                    if c >= self.map_height or c < 0 : continue
                    if r >= self.map_width or r < 0 : continue
                    vis_trajectory_map[r][c] = color_index + 2



        fig, ax = plt.subplots()
        ax.legend(self.trajectory_yaml_list)
        barrr = ax.imshow(vis_trajectory_map,cmap=cmap)
        fig.savefig('trajectory.png')


    def read_ply(self):
        # read ply file
        pc = o3d.io.read_point_cloud(self.yaml_data['source'])
        points_array = np.asarray((pc.points))

        self.map_height, self.map_width = self.yaml_data['map_height'], self.yaml_data['map_width']
        self.map = np.zeros((self.map_height, self.map_width))
        self.origin_x, self.origin_y = self.yaml_data['origin_x'], self.yaml_data['origin_y']
        self.resolution = self.yaml_data['resolution']
        for i in range(len(points_array)):
            if points_array[i][2] < self.yaml_data['z_max'] : continue
            if points_array[i][2] > self.yaml_data['z_min'] : continue

            c = int((points_array[i][0] - self.origin_x)/self.resolution)
            r = int((points_array[i][1] - self.origin_y)/self.resolution)

            if c >= self.map_height or c < 0 : continue
            if r >= self.map_width or r < 0 : continue

            self.map[r][c] += 1

        self.map_bit = np.zeros(self.map.shape)
        for c in range(self.map_height):
            for r in range(self.map_width):
                self.map_bit[r][c] = 1 if self.map[r][c] > self.yaml_data['threshold'] else 0

#         np.save(self.yaml_data['source'][:-4]+'_bit.npy', self.map_bit)
#         plt.imshow(self.map_bit)
#         plt.colorbar()
#         plt.show()

    def plot_metrics(self):
        load_obstacle_map = self.map_bit  #np.load(self.yaml_data['source'][:-4]+'_bit.npy')
        shape = load_obstacle_map.shape
        diff = DifficultyMetrics(load_obstacle_map, [], disp_radius=1)
        self.total_obstacle_map = load_obstacle_map
        self.total_closest_wall = diff.closest_wall()
        self.total_avg_visibility = diff.avg_visibility()
        self.total_dispersion = diff.dispersion()
        self.total_characteristic_dimension = diff.characteristic_dimension()


        str_list = ["obstacle_map",\
                    "closest_wall",\
                    "avg_visibility",\
                    "dispersion",\
                    "characteristic_dimension"]
        cmap_list = ["binary","RdYlGn","RdYlGn","RdYlGn","Greys"]
        for i,item in enumerate([self.total_obstacle_map,self.total_closest_wall,self.total_avg_visibility, \
                                 self.total_dispersion,self.total_characteristic_dimension]):
            fig, ax = plt.subplots()
            barrr = ax.imshow(item,cmap=cmap_list[i], interpolation='nearest')
            ax.set_title(str_list[i])
            fig.colorbar(barrr, ax=ax)
            fig.savefig(self.yaml_data['source'][:-4]+str_list[i]+".png")
            np.save(self.yaml_data['source'][:-4]+str_list[i]+".npy", np.asarray(item))


# AUTOGENERATED! DO NOT EDIT! File to edit: 10_gym_environment.ipynb (unless otherwise specified).

__all__ = ['show_env']

# Cell
import os
import re

def show_env(path, output_file):
    """

    This function will show all env on the path.

    Args:
        path : gym register file path.
        output_file :　The file to store all env name.

    Return:
        env : List all env name from the path.

    """

    os.system("cat " + path + " | grep \"id=\" >> " + output_file)

    env = []
    f = open(output_file, 'r')
    for line in f.readlines():
      if(line.find('#')):
        flag = [_.start() for _ in re.finditer("\'",line)]
        env.append(line[flag[0]+1:flag[1]])
    f.close()

    return env


# AUTOGENERATED! DO NOT EDIT! File to edit: 07_robot_model.ipynb (unless otherwise specified).

__all__ = ['show_link', 'show_joint']

# Cell

import gdown
from zipfile import ZipFile
import xml.etree.cElementTree as ET
import os

def show_link(path, output_file):
    """

    This function will read urdf.xacro file and show all link.

    Args :
        path : The urdf.xacro file path
        output_file : The file to store all link
    Return :
        link : list all link at the urdf.xacro file.
    """

    os.system("cat " + path + " | grep \"link name=\" >> " + output_file)

    link = []
    f = open(output_file, 'r')
    for line in f.readlines():
      index = []
      flag = 0
      for i, element in enumerate(line):
        if element == "<":
          flag = 1
        if flag == 1:
          index.append(i)
      link_string = str(line)
      link.append(link_string[index[0]:])
    f.close()

    return link

def show_joint(path):
    """

    This function will read xml file and show all joint from this file.

    Args:
        path : The xml file path
    Return:
        joint : list all joint at xml file.

    """

    joint = []
    tree = ET.parse(path)
    root = tree.getroot()

    for joints in root.iter("joint"):
        joint.append(joints.attrib)

    return joint


# AUTOGENERATED! DO NOT EDIT! File to edit: 05_DoorGym_gazebo_inference.ipynb (unless otherwise specified).

__all__ = ['download_model', 'init_model', 'inference']

# Cell
import torch
import os
import gdown
import yaml
from zipfile import ZipFile

# Cell
def download_model(id, path, name):
    """Download a model.

    This function will download model from google cloud.

    Args:
        id : The file id of google cloud.
        path : The path where download model.
        name : The output file name.

    Return:
        model_path : The model path.

    """
    dataset_url = 'https://drive.google.com/u/1/uc?id=' + id
    dataset_name = name + ".pt"
    model_path = os.path.join(path, "model", dataset_name)

    if not os.path.isdir(model_path):
        gdown.download(dataset_url, output = model_path, quiet=False)

    return model_path

# Cell
def init_model(model_path, state_dim):
    """Initial model.

    This function will load model and set need parameters.

    Args:
        model_path : The trained model path.
        state_dim : Dimension of state.

    Return：
        model : After initialization model.

    """
    actor_critic, ob_rms = torch.load(model_path)
    actor_critic = actor_critic.eval()
    actor_critic.nn = state_dim

    return actor_critic

# Cell
def inference(model, state, hidden_state):
    """Inference RL_mm in gazebo or real world.

    This function will use model and state to output next action.

    Args:
        model : The model after initalization.
        state : The environment and robot observation.
        hidden_state : The recurrent network setting.

    Return:
        action : The robot action.
        recurrent_hidden_state : The next hidden_state.

    """
    masks = torch.zeros(1, 1)
    with torch.no_grad():
        _, action, _, recurrent_hidden_states = model.act(
                        state, hidden_state, masks, deterministic=True)

    return action, recurrent_hidden_states


# AUTOGENERATED! DO NOT EDIT! File to edit: 08_environment_model.ipynb (unless otherwise specified).

__all__ = ['plot_obstacle_map', 'plot_trajectory', 'load_csv', 'load_world']

# Cell
import open3d as o3d
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap, LinearSegmentedColormap
import pylab as plot
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import yaml
import math
import os
from queue import Queue, PriorityQueue
import random
plt.rcParams["figure.figsize"] = (16, 16)

# Cell
def plot_obstacle_map(data_root, env, folder, filename):
    '''
        Plot obstacle map.
        Args:
            data_root (String) : The path of folder env_exp_files. Default at 'Data/env_exp_files' .
            env (String) : Which environment was used.
            folder (String) : The relative path to find the map file.
            filename (String) : The file name of numpy array map.

        Returns:
            obstacle_map (array) : obstacle map would also be saved in .npy files and visualized in .png files.

    '''
    map_path = data_root + '/' + env + '/' + folder + '/' + filename
    load_obstacle_map = np.load(map_path)
    obstacle_map = load_obstacle_map

    fig, ax = plt.subplots()
    barrr = ax.imshow(obstacle_map,cmap="binary", interpolation='nearest')
    ax.set_title("obstacle_map")
    fig.colorbar(barrr, ax=ax)
    npy_file_name = data_root + '/' + env + '/npy/' + "obstacle_map"+".npy"
    png_name = data_root + '/' + env + '/figure/' + "obstacle_map"+".png"
    fig.savefig(png_name)
    np.save(npy_file_name, np.asarray(load_obstacle_map))
    return obstacle_map

# Cell
def plot_trajectory(data_root, env, folder_map, map_file, folder_path, folder_info, info_file):
    """
        Calculate metrics for the input map.
        Args:
            data_root (String) : The path of folder env_exp_files. Default at 'Data/env_exp_files' .
            env (String) : Which environment was used.
            folder_map (String) : The relative path to find the yaml file `bit.map`.
            map_file (String) : The file name of the 2D occupency grid map.
            folder_path (String) : The relative path to find the yaml file of trajectories.
            folder_info (String) : The relative path to find the yaml file of information about 2D occupency grid map.
            info_file (String) : The file name of the yaml file about 2D occupency grid map.

        Returns:
            A figure (trajectory.png) would be saved.
    """

    yaml_path = data_root + '/' + env + '/' + folder_info + '/' + info_file
    with open(yaml_path, 'r') as f:
            info_yaml_data = yaml.load(f, Loader=yaml.CLoader)

    map_path = data_root + '/' + env + '/' + folder_map + '/' + map_file
    vis_trajectory_map = np.load(map_path)

    trajectory_folder = data_root + '/' + env + '/' + folder_path
    trajectory_yaml_list = os.listdir(trajectory_folder)

    cmap = ListedColormap(['white', 'black', 'red', 'green'])
    legend = []
    for index,baseline in enumerate(trajectory_yaml_list):
        legend.append(mpatches.Patch(color=['white', 'black', 'red', 'green'][index+2], label=baseline))

        baseline_path = trajectory_folder + '/' + baseline
        with open(baseline_path, 'r') as stream:
            yaml_data = yaml.load(stream, Loader=yaml.CLoader)

        for path in yaml_data["trajectories"]:
            for pose in path:
                c = int((pose["position"][0] - info_yaml_data["origin_x"])/info_yaml_data["resolution"])
                r = int((pose["position"][1] - info_yaml_data["origin_y"])/info_yaml_data["resolution"])

                if c >= info_yaml_data["map_width"] or c < 0 : continue
                if r >= info_yaml_data["map_height"] or r < 0 : continue
                vis_trajectory_map[r][c] = index + 2

    fig, ax = plt.subplots()
    ax.imshow(vis_trajectory_map)
    ax.legend(handles=legend, prop={'size': 16})
    barrr = ax.imshow(vis_trajectory_map,cmap=cmap)

    figure_path = data_root + '/' + env + '/' + 'figure'
    if not os.path.exists(figure_path):
        os.makedirs(os.path.join(data_root + '/' + env, 'figure'))

    fig.savefig(figure_path + '/trajectory.png')

# Cell
def load_csv(env, csv_file):
    '''
    load csv file into numpy array.
    Args :
        env (String) : Which environment was used.
        csv_files (String) : csv file path.

    Returns :
        array : pair points in numpy array.
    '''
    path = 'Data/env_exp_files/'+ env + '/environment_info/'+ csv_file
    with open(path) as file_name:
        array = np.loadtxt(file_name, delimiter=",")
    return array

# Cell
from os import listdir
from os.path import isfile, join

# Cell
def load_world(path):
    have_found=[]
    with open(path,'r') as f:
        for line in f.readlines():
            if ("model://" in line) :
                if not line in have_found:
                    print(line)
                    have_found.append(line)
        f.close()


# AUTOGENERATED! DO NOT EDIT! File to edit: 12_CurriculumSchedulerMultiMachine.ipynb (unless otherwise specified).

__all__ = ['CurriculumSchedulerMultiMachine']

# Cell
import os
import sys
import signal
import subprocess
import time
import numpy as np
import yaml
import random

# Cell
class CurriculumSchedulerMultiMachine():
	def __init__(self, yaml_path):
		with open(yaml_path, 'r') as f:
			self.yaml_data = yaml.load(f, Loader=yaml.CLoader)
			self.yaml_sort = sorted(self.yaml_data)
		self.success_queue = []
		self.current_task_index = 0
		self.current_task_env = ''
		self.process = None
		self.__switch_task()

	def update(self, success):
		"""
		Use a buffer and calculate average success rate for performance measure.
		You should update the buffer when the RL trainging episode `reset`.

		Args:
			success (Int) : $1$ if success and $0$ if fail.
		Returns:
			start_goal_pair (Dictionary) : A dictionary about the x,y,z position of start point and goal point.
										   i.e. ""{'goal': [33.65625, -7.6171875, 0.12646484375], 'start': [16.5, 40.40625, 0.12646484375]}""
			action_scale (Dictionary) : A dictionary about the output action scale.
										i.e. ""{'angular': 0.18, 'linear': 0.33, 'ur5': 0.5}"".
		"""
		current_task = self.yaml_sort[self.current_task_index]
		current_task_env = self.yaml_data[self.yaml_sort[self.current_task_index]]["environment"]
		window_size = self.yaml_data[current_task]["window_size"]
		threshold = self.yaml_data[current_task]["threshold"]
		self.success_queue.append(success)
		rate = np.mean(self.success_queue)
		print("------ CurriculumSchedulerMultiMachine ------")
		print("Current Task:",current_task,current_task_env)
		print("window_size:",window_size, ", Current Buffer size:",len(self.success_queue))
		print("Threshold:",threshold, ", Current Avg success rate:",rate)

		if len(self.success_queue) >= window_size:
			self.success_queue = self.success_queue[-window_size:]
			if (self.current_task_index<len(self.yaml_sort)-1) and rate>threshold:
				self.current_task_index += 1
				self.success_queue = []
				self.__switch_task()
		start_goal_pair = self.yaml_data[current_task]['pairs'][random.randint(0,len(self.yaml_data[current_task]['pairs'])-1)]
		action_scale = self.yaml_data[current_task]['action_scale']
		return start_goal_pair, action_scale

	def __switch_task(self):
		"""
		A flag for Gym to decide which socket package should be received.
		"""
		self.current_task_env = self.yaml_data[self.yaml_sort[self.current_task_index]]["environment"]



# AUTOGENERATED! DO NOT EDIT! File to edit: 09_CurriculumSchedulerOneMachine.ipynb (unless otherwise specified).

__all__ = ['CurriculumSchedulerOneMachine', 'train_env_specify']

# Cell
import os
import sys
import signal
import subprocess
import time
import numpy as np
import yaml
import random

# Cell
class CurriculumSchedulerOneMachine():
	def __init__(self, yaml_path, command_dic, waiting_second=40):
		"""
		Curriculum Scheduler controls the process of training with multi tasks.
		Args:
			yaml_path (String) : The path of an yaml file which includes all information about task,
								 such as environment, start and goal points pairs, output action scale, threshold ($c$) and window_size ($k$).
			command_dic (Dictionary) : A dictionary maps environment to the `shell command` using to launch the simulation environment.
			waiting_second (Int) : The number of seconds to wait when switching simulation environments.
		"""
		self.command_dic = command_dic
		self.waiting_second = waiting_second
		with open(yaml_path, 'r') as f:
			self.yaml_data = yaml.load(f, Loader=yaml.CLoader)
			self.yaml_sort = sorted(self.yaml_data)
		self.success_queue = []
		self.current_task_index = 0
		self.process = None
		self.__switch_task()

	def update(self, success):
		"""
		Use a buffer and calculate average success rate for performance measure.
		You should update the buffer when the RL trainging episode `reset`.

		Args:
			success (Int) : $1$ if success and $0$ if fail.
		Returns:
			start_goal_pair (Dictionary) : A dictionary about the x,y,z position of start point and goal point.
										   i.e. ""{'goal': [33.65625, -7.6171875, 0.12646484375], 'start': [16.5, 40.40625, 0.12646484375]}""
			action_scale (Dictionary) : A dictionary about the output action scale.
										i.e. ""{'angular': 0.18, 'linear': 0.33, 'ur5': 0.5}"".
		"""
		current_task = self.yaml_sort[self.current_task_index]
		current_task_env = self.yaml_data[self.yaml_sort[self.current_task_index]]["environment"]
		window_size = self.yaml_data[current_task]["window_size"]
		threshold = self.yaml_data[current_task]["threshold"]
		self.success_queue.append(success)
		rate = np.mean(self.success_queue)
		print("------ CurriculumSchedulerOneMachine ------")
		print("Current Task:",current_task,current_task_env)
		print("window_size:",window_size, ", Current Buffer size:",len(self.success_queue))
		print("Threshold:",threshold, ", Current Avg success rate:",rate)

		if len(self.success_queue) >= window_size:
			self.success_queue = self.success_queue[-window_size:]
			if (self.current_task_index<len(self.yaml_sort)-1) and rate>threshold:
				self.current_task_index += 1
				self.success_queue = []
				self.__switch_task()
		start_goal_pair = self.yaml_data[current_task]['pairs'][random.randint(0,len(self.yaml_data[current_task]['pairs'])-1)]
		action_scale = self.yaml_data[current_task]['action_scale']
		return start_goal_pair, action_scale

	def __switch_task(self):
		"""
		Close the old simulation evironment and launch a new one.
		"""
		if not self.process==None:
			os.killpg(os.getpgid(self.process.pid), signal.SIGTERM)  # Send the signal to all the process groups
			for i in range(self.waiting_second):
				time.sleep(1)
				print("waiting", self.waiting_second-i)

		current_task_env = self.yaml_data[self.yaml_sort[self.current_task_index]]["environment"]
		command_str = self.command_dic[current_task_env]
		self.process = subprocess.Popen(command_str, stdout=subprocess.PIPE,shell=True, preexec_fn=os.setsid)
		print('\033[92m@@@@@@@@@ \\\ Level Up /// @@@@@@@@@\033[0m')
		for i in range(self.waiting_second):
			time.sleep(1)
			print("waiting", self.waiting_second-i)


# Cell
def train_env_specify(task_list, data_root='Data'):
	"""
		Automatically generate a well formatted yaml file for the specific environment.

		Args:
			task_list (List) : A List of environment name (String).
							   i.e. ["cave","forest","campus"]
		Returns:
			A yaml file named `curriculum.yaml` would be save at data_root folder.
		"""
	assert os.path.exists(data_root),"data_root Path not found"
	for task in task_list:
		assert os.path.exists(data_root+'/'+task+'.yaml'),"No environment yaml file"

	CL_training = dict()
	for index,task in enumerate(task_list):
		with open(data_root+'/'+task+'.yaml', 'r') as f:
			value = yaml.load(f, Loader=yaml.CLoader)
		key = 'task' + (str(index+1)).zfill(2)
		CL_training[key] = value

	with open(data_root+'/'+'curriculum.yaml', 'w') as f:
		yaml.dump(CL_training, f)


# AUTOGENERATED! DO NOT EDIT! File to edit: 01_model_utils.ipynb (unless otherwise specified).

__all__ = ['download_model', 'load_model', 'inference']

# Cell
def download_model(id, name):
    """Download a model.

    This function will download model from google drive.

    Args:
        id: The file id of google cloud.
        name: The output zip file name.

    Return:
        path: The zip model path.

    """
    dataset_url = 'https://drive.google.com/u/1/uc?id=' + id
    dataset_name = name + ".zip"
    path = os.path.join(os.getcwd(), "data", dataset_name)

    if not os.path.isdir(path):
        gdown.download(dataset_url, output = path, quiet=False)

    return path
def load_model(model_path):
    """Load the policy model.
    Args:
        model_path: The path of the policy weight.
    Return:
        path: model(policy network).
    """
    gpu = tf.config.experimental.list_physical_devices('GPU')
    tf.config.experimental.set_memory_growth(gpu[0], True)
    policy_network = tf.saved_model.load(model_path)
    return policy_network

def inference(model, state):
    """ Inference RL_vrx in gazebo or real world
    This function will use policy model and state to output the next action
    Args:
        model: policy model.
        state: observation.
    Return:
        action : The robot action[v,w].
    """
    state = tf.convert_to_tensor([state], dtype=tf.float32)
    action = model(state)[0].numpy()
    return action



# AUTOGENERATED! DO NOT EDIT! File to edit: 03_diff_visualization.ipynb (unless otherwise specified).

__all__ = ['DifficultyMetrics', 'read_ply', 'plot_metric', 'plot_trajectory', 'avg_distance']

# Cell
import open3d as o3d
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap, LinearSegmentedColormap
import pylab as plot
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import yaml
import math
import os
from queue import Queue, PriorityQueue
plt.rcParams["figure.figsize"] = (16, 16)

# Cell
class DifficultyMetrics:
      def __init__(self):
        pass

      def closest_wall(self, input_map):
        """
        Calculate the avg Distance to Closest Obstacle of an occupancy grid map.
        Args:
            input_map (2D array) : occupancy grid map.
        Returns:
            Avg Distance to Closest Obstacle of all cell in map.
        """
        dists = [[0 for i in range(len(input_map[0]))] for j in range(len(input_map))]
        for r in range(len(input_map)):
          for c in range(len(input_map[0])):
            dists[r][c] = self._dist_closest_wall(r, c, input_map)

        return dists


      def avg_visibility(self, input_map):
        """
        Calculate the average visibility of an occupancy grid map.
        Args:
            input_map (2D array) : occupancy grid map.
        Returns:
            The average visibility at each point.
        """
        vis = [[0 for i in range(len(input_map[0]))] for j in range(len(input_map))]
        for r in range(len(input_map)):
          for c in range(len(input_map[0])):
            vis[r][c] = self._avg_vis_cell(r, c, input_map)

        return vis


      def dispersion(self, input_map):
        """
        Calculate the average dispersion of an occupancy grid map.
        Checks along 16 axes within the dispersion radius.
        Args:
            input_map (2D array) : occupancy grid map.
        Returns:
            The average dispersion at each point.
        """
        disp = [[0 for i in range(len(input_map[0]))] for j in range(len(input_map))]
        for r in range(len(input_map)):
          for c in range(len(input_map[0])):
            disp[r][c] = self._cell_dispersion(r, c, input_map)

        return disp


      def characteristic_dimension(self,input_map):
        """
        Calculate the average dispersion of an occupancy grid map.
        Characteristic dimension calculated in 2 directions for 4 axes
        Args:
            input_map (2D array) : occupancy grid map.
        Returns:
            The average of characteristic dimension at each point.
        """
        cdr = [[0 for i in range(len(input_map[0]))] for j in range(len(input_map))]
        for r in range(len(input_map)):
          for c in range(len(input_map[0])):
            if input_map[r][c] == 1:
              cdr[r][c] = 0

            cdr_min = len(input_map) + len(input_map[0])
            for axis in [(0, 1), (1, 1), (1, 0), (1, -1)]:
              cdr_min = min(cdr_min, self._distance(r, c, axis, input_map))

            cdr[r][c] = cdr_min

        return cdr


      def _distance(self, r, c, axis, input_map):
        """
        Args:
            input_map (2D array) : occupancy grid map.
            r(int) : index in row.
            c(int) : index in column.
            axis(int) : a list such as [0,1], indicate the directions.
        Returns:
            The distance along the axis in both directions, not including (r, c)
        """
        if input_map[r][c] == 1:
          return -1

        # check axis in both directions
        reverse_axis = (axis[0] * -1, axis[1] * -1)
        dist = 0
        for move in [axis, reverse_axis]:
          r_curr = r
          c_curr = c

          # move in axis direction until an obstacle is found
          while input_map[r_curr][c_curr] != 1:
            r_curr += move[0]
            c_curr += move[1]

            if not self._in_map(r_curr, c_curr, input_map):
              break

            # add the distance traveled to the total
            if input_map[r_curr][c_curr] != 1:
              dist += math.sqrt(move[0] ** 2 + move[1] ** 2)

        return dist



      def _dist_between_points(self, a, b):
        """
        Args:
            a and b are tuples : (row, col)
        Returns:
            The distance between points
        """
        return math.sqrt((a[0] - b[0]) ** 2 + (a[1] - b[1]) ** 2)



      def _cell_dispersion(self, r, c, input_map, radius=1):
        """
        Args:
            input_map (2D array) : occupancy grid map.
            r(int) : index in row.
            c(int) : index in column.
            radius(int) : cell size, default as 1.
        Returns:
            The cell dispersion within the radius.
        """
        if input_map[r][c] == 1:
          return -1

        axes_wall = []
        # four cardinal, four diagonal, and one in between each (slope +- 1/2 or +-2)
        for move in [(0, 1), (1, 2), (1, 1), (2, 1), (1, 0), (2, -1), (1, -1), (1, -2), (0, -1), (-2, -1), (-1, -1), (-1, -2), (-1, 0), (-2, 1), (-1, 1), (-1, 2)]:
          count = 0
          wall = False
          r_curr = r
          c_curr = c
          while count < radius and not wall:
            r_curr += move[0]
            c_curr += move[1]

            if r_curr < 0 or r_curr >= len(input_map) or c_curr < 0 or c_curr >= len(input_map[0]):
              break

            if input_map[r_curr][c_curr] == 1:
              wall = True

            # count the in-between axes as two steps
            if move[0] == 2 or move[1] == 2:
              count += 2
            else:
              count += 1

          if wall:
            axes_wall.append(True)
          else:
            axes_wall.append(False)

        # count the number of changes in this cell's field of view
        change_count = 0
        for i in range(len(axes_wall)-1):
          if axes_wall[i] != axes_wall[i+1]:
            change_count += 1

        if axes_wall[0] != axes_wall[15]:
          change_count += 1

        return change_count


      def _avg_vis_cell(self, r, c, input_map):
        """
        Args:
            input_map (2D array) : occupancy grid map.
            r(int) : index in row.
            c(int) : index in column.
        Returns:
            The average visibility at the point (r, c).
        """
        total_vis = 0.0
        num_axes = 0
        for r_move in [-1, 0, 1]:
          for c_move in [-1, 0, 1]:
            if r_move == 0 and c_move == 0:
              continue

            this_vis = 0
            r_curr = r
            c_curr = c
            wall_found = False
            while not wall_found:
              if not self._in_map(r_curr, c_curr, input_map):
                break

              if input_map[r_curr][c_curr] == 1:
                wall_found = True
              else:
                this_vis += math.sqrt((r_move ** 2) + (c_move ** 2))

              r_curr += r_move
              c_curr += c_move

            total_vis += this_vis
            num_axes += 1

        return total_vis / num_axes



      def _in_map(self, r, c, input_map):
        """
        Bounds check.
        Args:
            r(int) : index in row.
            c(int) : index in column.
        Returns:
            A Boolean means that if the cell is in bounded.
        """
        return r >= 0 and r < len(input_map) and c >= 0 and c < len(input_map[0])


      def _dist_closest_wall(self, r, c, input_map):
        """
        Args:
            input_map (2D array) : occupancy grid map.
            r(int) : index in row.
            c(int) : index in column.
        Returns:
            The distance to the closest obstacle at point (r, c).
            Returns 0 if input_map[r][c] is an obstacle, 1 if an adjacent non-diagonal cell is an obstacle, etc.
        """
        pq = PriorityQueue()
        first_wrapper = self.Wrapper(0, r, c)
        pq.put(first_wrapper)
        visited = {(r, c) : first_wrapper}

        while not pq.empty():
          point = pq.get()
          if input_map[point.r][point.c] == 1: # found an obstacle!
            return point.dist
          else:
            # enqueue all neighbors if they are in the map and have not been visited
            for row in range(point.r - 1, point.r + 2):
              for col in range(point.c - 1, point.c + 2):
                if self._in_map(row, col, input_map) and (row, col) not in visited:
                  dist = math.sqrt((row - r) ** 2 + (col - c) ** 2)
                  neighbor = self.Wrapper(dist, row, col)
                  pq.put(neighbor)
                  visited[(row, col)] = neighbor

        # in case the queue is empty before a wall is found (shouldn't happen),
        # the farthest a cell can be from a wall is half the board, since the top and bottom rows are all walls
        return (len(input_map) - 1) / 2


      class Wrapper:
        """ wrapper class for coordinates """
        def __init__(self, distance, row, col):
          self.dist = distance
          self.r = row
          self.c = col

        def __lt__(self, value):
          return self.dist < value.dist


# Cell

def read_ply(data_root, env, folder, filename):
    """
        Generate the 2D occupency grid map by 3D pointcloud map.
        Args:
            data_root (String) : The path of folder env_exp_files. Default at 'Data/env_exp_files' .
            env (String) : Which environment was used.
            folder (String) : The relative path to find the yaml file.
            filename (String) : The file name of yaml.

            Finally, the full path of the target file would be "data_root + '/' + env + '/' + folder + '/' + filename".
        Returns:
            map_bit (array) : The occupency grid map in the form of an numpy array. It would also be saved in .npy file and visualized in .png file.
    """
    # read ply file
    yaml_path = data_root + '/' + env + '/' + folder + '/' + filename
    print(yaml_path)
    with open(yaml_path, 'r') as stream:
        yaml_data = yaml.load(stream, Loader=yaml.CLoader)
    pc = o3d.io.read_point_cloud(data_root + '/' + env + '/' + folder + '/' + yaml_data['source'])

    points_array = np.asarray((pc.points))

    map_height, map_width = yaml_data['map_height'], yaml_data['map_width']
    map = np.zeros((map_height, map_width))
    origin_x, origin_y = yaml_data['origin_x'], yaml_data['origin_y']
    resolution = yaml_data['resolution']
    for i in range(len(points_array)):
        if points_array[i][2] < yaml_data['z_max'] : continue
        if points_array[i][2] > yaml_data['z_min'] : continue

        c = int((points_array[i][0] - origin_x)/resolution)
        r = int((points_array[i][1] - origin_y)/resolution)

        if c >= map_width or c < 0 :
            continue
        if r >= map_height or r < 0 :
            continue

        map[r][c] += 1

    map_bit = np.zeros(map.shape)
    for c in range(map_width):
        for r in range(map_height):
            map_bit[r][c] = 1 if map[r][c] > yaml_data['threshold'] else 0

    folder_path = os.path.join(data_root + '/' + env, 'npy')
    if not os.path.exists(folder_path):
        os.makedirs(os.path.join(data_root + '/' + env, 'npy'))

    npy_file_name = data_root + '/' + env + '/npy/' + 'bit.npy'
    np.save(npy_file_name, map_bit)
    plt.imshow(map_bit)
    plt.colorbar()
    plt.show()
    return map_bit


# Cell

def plot_metric(data_root, env, folder, filename):
    """
        Calculate metrics for the input map.
        Args:
            data_root (String) : The path of folder env_exp_files. Default at 'Data/env_exp_files' .
            env (String) : Which environment was used.
            folder (String) : The relative path to find the yaml file.
            filename (String) : The file name of yaml.

            Finally, the full path of the .npy file for the occupency grid map would be "data_root + '/' + env + '/' + folder + '/' + filename".

        Returns:
            5 map_bit (array) : The result of metrics which order in [`obstacle map`,`closest wall`,`visibility`, `dispersion`,`characteristic dimension`].
                                They would also be saved in .npy files and visualized in .png files.
    """
    map_path = data_root + '/' + env + '/' + folder + '/' + filename
    load_obstacle_map = np.load(map_path)
    shape = load_obstacle_map.shape
    diff = DifficultyMetrics()
    total_obstacle_map = load_obstacle_map
    total_closest_wall = diff.closest_wall(load_obstacle_map)
    total_avg_visibility = diff.avg_visibility(load_obstacle_map)
    total_dispersion = diff.dispersion(load_obstacle_map)
    total_characteristic_dimension = diff.characteristic_dimension(load_obstacle_map)

    str_list = ["obstacle_map",\
                "closest_wall",\
                "avg_visibility",\
                "dispersion",\
                "characteristic_dimension"]
    cmap_list = ["binary","RdYlGn","RdYlGn","RdYlGn","Greys"]

    folder_path = os.path.join(data_root + '/' + env, 'npy')
    if not os.path.exists(folder_path):
        os.makedirs(os.path.join(data_root + '/' + env, 'npy'))
    figure_path = os.path.join(data_root + '/' + env, 'figure')
    if not os.path.exists(figure_path):
        os.makedirs(os.path.join(data_root + '/' + env, 'figure'))


    for i,item in enumerate([total_obstacle_map,total_closest_wall,total_avg_visibility, \
                             total_dispersion,total_characteristic_dimension]):
        fig, ax = plt.subplots()
        barrr = ax.imshow(item,cmap=cmap_list[i], interpolation='nearest')
        ax.set_title(str_list[i])
        fig.colorbar(barrr, ax=ax)
        npy_file_name = data_root + '/' + env + '/npy/' + str_list[i]+".npy"
        png_name = data_root + '/' + env + '/figure/' + str_list[i]+".png"
        fig.savefig(png_name)
        np.save(npy_file_name, np.asarray(item))

    return total_obstacle_map, total_closest_wall, total_avg_visibility, total_dispersion, total_characteristic_dimension

# Cell
def plot_trajectory(data_root, env, folder_map, map_file, folder_path, folder_info, info_file):
    """
        Calculate metrics for the input map.
        Args:
            data_root (String) : The path of folder env_exp_files. Default at 'Data/env_exp_files' .
            env (String) : Which environment was used.
            folder_map (String) : The relative path to find the yaml file `bit.map`.
            map_file (String) : The file name of the 2D occupency grid map.
            folder_path (String) : The relative path to find the yaml file of trajectories.
            folder_info (String) : The relative path to find the yaml file of information about 2D occupency grid map.
            info_file (String) : The file name of the yaml file about 2D occupency grid map.

        Returns:
            A figure (trajectory.png) would be saved.
    """

    yaml_path = data_root + '/' + env + '/' + folder_info + '/' + info_file
    with open(yaml_path, 'r') as f:
            info_yaml_data = yaml.load(f, Loader=yaml.CLoader)

    map_path = data_root + '/' + env + '/' + folder_map + '/' + map_file
    vis_trajectory_map = np.load(map_path)

    trajectory_folder = data_root + '/' + env + '/' + folder_path
    trajectory_yaml_list = os.listdir(trajectory_folder)

    cmap = ListedColormap(['white', 'black', 'red', 'green'])
    legend = []
    for index,baseline in enumerate(trajectory_yaml_list):
        legend.append(mpatches.Patch(color=['white', 'black', 'red', 'green'][index+2], label=baseline))

        baseline_path = trajectory_folder + '/' + baseline
        with open(baseline_path, 'r') as stream:
            yaml_data = yaml.load(stream, Loader=yaml.CLoader)

        for path in yaml_data["trajectories"]:
            for pose in path:
                c = int((pose["position"][0] - info_yaml_data["origin_x"])/info_yaml_data["resolution"])
                r = int((pose["position"][1] - info_yaml_data["origin_y"])/info_yaml_data["resolution"])

                if c >= info_yaml_data["map_width"] or c < 0 : continue
                if r >= info_yaml_data["map_height"] or r < 0 : continue
                vis_trajectory_map[r][c] = index + 2



    fig, ax = plt.subplots()
    ax.imshow(vis_trajectory_map)
    ax.legend(handles=legend, prop={'size': 16})
    barrr = ax.imshow(vis_trajectory_map,cmap=cmap)

    figure_path = data_root + '/' + env + '/' + 'figure'
    if not os.path.exists(figure_path):
        os.makedirs(os.path.join(data_root + '/' + env, 'figure'))

    fig.savefig(figure_path + '/trajectory.png')

# Cell
def avg_distance(data_root, env, folder_info, csv_file, info_file):
    """
        Calculate metrics for the input map.
        Args:
            data_root (String) : The path of folder env_exp_files. Default at 'Data/env_exp_files' .
            env (String) : Which environment was used.
            folder_info (String) : The relative path to find the yaml file of information about 2D occupency grid map.
            csv_file (String) : The file name of the csv file about Start and Goal points.

        Returns:
            avg_dis (Float) : The average of the distance between Start and Goal points.
            map_width (Int) : Map width (meter).
            map_height (Int) : Map width (meter).
    """

    csv_path = data_root + '/' + env + '/' + folder_info + '/' + csv_file
    csv_data = np.genfromtxt(csv_path, delimiter=',')

    yaml_path = data_root + '/' + env + '/' + folder_info + '/' + info_file
    with open(yaml_path, 'r') as f:
        info_yaml_data = yaml.load(f, Loader=yaml.CLoader)

    total_dis = 0
    for i in range(len(csv_data)):
        if i%2==1: continue
        x1, y1 = csv_data[i][0], csv_data[i][1]
        x2, y2 = csv_data[i+1][0], csv_data[i+1][1]

        dis = ((x2-x1)**2 + (y2-y1)**2) ** 0.5
        total_dis += dis

    avg = 2*total_dis/len(csv_data)

    print("----------",env,"----------")
    print("Start and Goal points pairs: ", len(csv_data)/2)
    print("The average of the distance between Start and Goal points: ",avg)
    map_width = info_yaml_data["map_width"]*info_yaml_data["resolution"]
    map_height = info_yaml_data["map_height"]*info_yaml_data["resolution"]
    print("Map Width: ",map_width)
    print("Map Height: ",map_height)

    return avg, map_width, map_height
